#include "QVistaMicrophone.h"
#include <stdio.h>

QVistaMicrophone::QVistaMicrophone(BOOL bCalibration, UINT uChunk)
{
	SYSTEMTIME st;

	bCalibrate = bCalibration;
	bSilence = FALSE;
	uChunkLen = uChunk;
	uAudioBufferSize = 0;
	uChannels = 0;
	pAudioBuffer = NULL;

	wfx.wFormatTag = WAVE_FORMAT_PCM;
	wfx.nSamplesPerSec = 44100;
	wfx.nChannels = 2;
	wfx.wBitsPerSample = 16;
	wfx.nBlockAlign = (wfx.nChannels * wfx.wBitsPerSample) / 8;
	wfx.nAvgBytesPerSec = wfx.nSamplesPerSec * wfx.nBlockAlign;
	wfx.cbSize = 0;

	pEnumerator = NULL;
	pDevice = NULL;
	pAudioClient = NULL;
	pAudio = NULL;
	pCaptureClient = NULL;

	GetSystemTime(&st);
	i64TimeStamp = (INT64)&st;

	hr = CoInitializeEx(NULL, COINIT_APARTMENTTHREADED);
	InitMic();
}

QVistaMicrophone::~QVistaMicrophone()
{
	if(pAudioBuffer != NULL){
		delete[] pAudioBuffer;
		pAudioBuffer = NULL;
		uAudioBufferSize = 0;
	}

	SAFE_RELEASE(pEnumerator)
	SAFE_RELEASE(pDevice)
	SAFE_RELEASE(pAudioClient)
	SAFE_RELEASE(pAudio)
	SAFE_RELEASE(pCaptureClient)

	CoUninitialize();
}

// Torna la dimensione del buffer allocato, va chiamata dopo l'AcquireMic()
UINT QVistaMicrophone::GetBufferSize()
{
	if(IsInitialized() == FALSE)
		return 0;

	return uAudioBufferSize;
}

BOOL QVistaMicrophone::AcquireMic(PBYTE* pBuffer)
{
	if(pBuffer == NULL)
		return FALSE;

	if(hr != S_OK && hr != S_FALSE)
		return FALSE;

	if(pAudioBuffer){
		delete[] pAudioBuffer;
		pAudioBuffer = NULL;
		uAudioBufferSize = 0;
	}

	if(IsInitialized() == FALSE){
		if(InitMic() == FALSE)
			return FALSE;
	}

	if(AcquireMic() == FALSE)
		return FALSE;

	*pBuffer = pAudioBuffer;
	return TRUE;
}

BOOL QVistaMicrophone::AcquireMic()
{
	BOOL bDone = FALSE, bAppend = FALSE;
	UINT32 bufferFrameCount, numFramesAvailable, packetLength = 0, uTmp = 0;
	UINT uSleepTime, uSleepSlice;
	PBYTE pData = NULL, pTmp = NULL;
	DWORD flags, dwSleep = 0;
	REFERENCE_TIME hnsRequestedDuration = REFTIMES_PER_SEC;
	REFERENCE_TIME hnsActualDuration;

	if(IsInitialized() == FALSE)
		return FALSE;

	CalibrateMic();

	// Get mute state
	hr = pAudio->GetMute(&bMute);
	EXIT_ON_ERROR(hr)

	if(bMute == TRUE){
		if(pAudioBuffer){
			delete[] pAudioBuffer;
			pAudioBuffer = NULL;
			uAudioBufferSize = 0;
		}

		return TRUE;
	}

	// Get the size of the allocated buffer.
	hr = pAudioClient->GetBufferSize(&bufferFrameCount);
	EXIT_ON_ERROR(hr)

	// Calculate the actual duration of the allocated buffer.
	hnsActualDuration = (REFERENCE_TIME)((double)(REFTIMES_PER_SEC) * bufferFrameCount / wfx.nSamplesPerSec);

	if(pAudioBuffer){
		delete[] pAudioBuffer;
		pAudioBuffer = NULL;
		uAudioBufferSize = 0;
	}

	// Calcoliamo il tempo effettivo che spendiamo nella sleep e nella copia
	// dei dati, in modo da riempire il piu' possibile il buffer
	uSleepTime = (UINT)(hnsActualDuration / REFTIMES_PER_MILLISEC);
	uSleepSlice = uSleepTime / 10;
	dwSleep = uSleepTime - (uSleepTime / 10); // Total sleep time - 10%
	dwSleep += GetTickCount() - uSleepSlice;

	// Start recording.
	hr = pAudioClient->Start();
	EXIT_ON_ERROR(hr)

	while(dwSleep > GetTickCount()){
		// Sleep for 1/10th of the buffer duration.
		MiniSleep(uSleepSlice);

		hr = pCaptureClient->GetNextPacketSize(&packetLength);
		EXIT_ON_ERROR(hr)
		
		while(packetLength != 0){
			// Get the available data in the shared buffer.
			hr = pCaptureClient->GetBuffer(&pData, &numFramesAvailable, &flags, NULL, NULL);
			EXIT_ON_ERROR(hr)

			if(flags & AUDCLNT_BUFFERFLAGS_SILENT){
				pData = NULL;
				return TRUE;
			}

			// The size, in bytes, of an audio frame equals the number 
			// of channels in the stream multiplied by the sample size per channel. 
			// For example, for a stereo (2-channel) stream with 16-bit samples, the 
			// frame size is four bytes.
			if(pData){
				if(pAudioBuffer){ // Estendiamo il buffer
					uTmp = numFramesAvailable * (wfx.nChannels * wfx.wBitsPerSample / 8);

					pTmp = new(std::nothrow) BYTE[uAudioBufferSize + uTmp + 1];

					if(pTmp == NULL){
						delete[] pAudioBuffer;
						pAudioBuffer = NULL;
						uAudioBufferSize = 0;
						return FALSE;
					}

					memcpy(pTmp, pAudioBuffer, uAudioBufferSize);
					memcpy(pTmp + uAudioBufferSize, pData, uTmp);

					delete[] pAudioBuffer;
					pAudioBuffer = pTmp;
					uAudioBufferSize += uTmp;
					pTmp = NULL;
				}else{ // Allochiamo il buffer
					uAudioBufferSize = numFramesAvailable * (wfx.nChannels * wfx.wBitsPerSample / 8);
					pAudioBuffer = new(std::nothrow) BYTE[uAudioBufferSize + 1];

					if(pAudioBuffer == NULL){
						uAudioBufferSize = 0;
						return FALSE;
					}

					memcpy(pAudioBuffer, pData, uAudioBufferSize);
				}
			}

			hr = pCaptureClient->ReleaseBuffer(numFramesAvailable);
			EXIT_ON_ERROR(hr)

			hr = pCaptureClient->GetNextPacketSize(&packetLength);
			EXIT_ON_ERROR(hr)
		}
	}

	// Stop recording.
	hr = pAudioClient->Stop();
	EXIT_ON_ERROR(hr)

	DetectSilence();
	return TRUE;

Exit:
	if(pAudioBuffer){
		delete[] pAudioBuffer;
		pAudioBuffer = NULL;
		uAudioBufferSize = 0;
	}

	return FALSE;
}

BOOL QVistaMicrophone::InitMic()
{
	REFERENCE_TIME hnsRequestedDuration = REFTIMES_PER_SEC;
	WAVEFORMATEXTENSIBLE wfxe;

	if(IsInitialized() == TRUE)
		return TRUE;

	if(CoCreateInstance(__uuidof(MMDeviceEnumerator), NULL, CLSCTX_ALL, __uuidof(IMMDeviceEnumerator), (LPVOID*)&pEnumerator) != S_OK)
		return FALSE;

	hr = pEnumerator->GetDefaultAudioEndpoint(eCapture, eConsole, &pDevice);
	EXIT_ON_ERROR(hr)

	hr = pDevice->Activate(__uuidof(IAudioClient), CLSCTX_ALL, NULL, (void**)&pAudioClient);
	EXIT_ON_ERROR(hr)

	hr = pDevice->Activate(__uuidof(IAudioEndpointVolume), CLSCTX_ALL, NULL, (void**)&pAudio);
	EXIT_ON_ERROR(hr)

	hr = pAudio->GetChannelCount(&uChannels);
	EXIT_ON_ERROR(hr)

	hr = pAudio->GetVolumeRange(&fMin, &fMax, &fStep);
	EXIT_ON_ERROR(hr)

	hr = pAudioClient->IsFormatSupported(AUDCLNT_SHAREMODE_SHARED, &wfx, (WAVEFORMATEX **)&wfxe);
	EXIT_ON_ERROR(hr)

	hr = pAudioClient->Initialize(AUDCLNT_SHAREMODE_SHARED, 0, REFTIMES_PER_SEC * uChunkLen, 0, &wfx, NULL);
	EXIT_ON_ERROR(hr)

	hr = pAudioClient->GetService(__uuidof(IAudioCaptureClient), (void**)&pCaptureClient); 
	EXIT_ON_ERROR(hr)

	bMicInitialization = TRUE;
	return bMicInitialization;

Exit:
	SAFE_RELEASE(pEnumerator)
	SAFE_RELEASE(pDevice)
	SAFE_RELEASE(pAudioClient)
	SAFE_RELEASE(pCaptureClient)
	bMicInitialization = FALSE;
	return bMicInitialization;
}

BOOL QVistaMicrophone::IsInitialized()
{
	return bMicInitialization;
}

void QVistaMicrophone::MiniSleep(UINT uMilliSec)
{
	UINT uLoops;
	UINT uSleepTime = 300; // Step di 300ms

	CANCELLATION_POINT(bAmbientalMicSemaphore)

	if(uMilliSec <= uSleepTime){
		Sleep(uMilliSec);
		return;
	}else{
		uLoops = uMilliSec / uSleepTime;
	}

	while(uLoops){
		Sleep(uSleepTime);
		uLoops--;
		CANCELLATION_POINT(bAmbientalMicSemaphore)
	}
}

BOOL QVistaMicrophone::CalibrateMic()
{
	HRESULT hr;
	UINT  i;
	FLOAT fLevel = 0.80f;
	IAudioEndpointVolume *pAudio = NULL;

	if(bCalibrate == FALSE)
		return FALSE;

	if(IsInitialized() == FALSE){
		if(InitMic() == FALSE)
			return FALSE;
	}

	// Set mute state to FALSE
	hr = pAudio->SetMute(FALSE, NULL);
	pAudio->GetMute(&bMute);

	// Torna FALSE se non riusciamo ad effettuare l'un-mute
	if(bMute == TRUE)
		return FALSE;

	if(uChannels == 0)
		return FALSE;

	// Normalizza all'80% tutti i canali in ingresso
	for(i = 0; i < uChannels; i++)
		pAudio->SetChannelVolumeLevelScalar(i, fLevel, NULL);

	return TRUE;
}

// Alza il guadagno del 10%
BOOL QVistaMicrophone::StepUp()
{
	FLOAT fLevel = 0.10f;
	UINT i;

	if(bCalibrate == FALSE)
		return FALSE;

	if(bMicInitialization == FALSE){
		if(InitMic() == FALSE)
			return FALSE;
	}

	// Set mute state to FALSE
	hr = pAudio->SetMute(FALSE, NULL);
	pAudio->GetMute(&bMute);

	// Torna FALSE se non riusciamo ad effettuare l'un-mute
	if(bMute == TRUE)
		return FALSE;

	if(uChannels == 0)
		return FALSE;

	for(i = 0; i < uChannels; i++){
		pAudio->GetChannelVolumeLevelScalar(i, &fLevel);
		pAudio->SetChannelVolumeLevelScalar(i, fLevel + 0.10f, NULL);
	}

	return TRUE;
}

// Abbassa il guadagano del 10%
BOOL QVistaMicrophone::StepDown()
{	
	FLOAT fLevel = 0.10f;
	UINT i;

	if(bCalibrate == FALSE)
		return FALSE;

	if(bMicInitialization == FALSE){
		if(InitMic() == FALSE)
			return FALSE;
	}

	// Set mute state to FALSE
	hr = pAudio->SetMute(FALSE, NULL);
	pAudio->GetMute(&bMute);

	// Torna FALSE se non riusciamo ad effettuare l'un-mute
	if(bMute == TRUE)
		return FALSE;

	if(uChannels == 0)
		return FALSE;

	for(i = 0; i < uChannels; i++){
		pAudio->GetChannelVolumeLevelScalar(i, &fLevel);
		pAudio->SetChannelVolumeLevelScalar(i, fLevel - 0.10f, NULL);
	}

	return TRUE;

}

// Se il campione acquisito contiene solo silenzio, cancella
// il buffer, setta la dimensione dei dati acquisiti a 0
// e setta bSilence a TRUE. Altrimenti setta bSilence a FALSE
// e lascia tutto invariato. Torna TRUE se il campione attuale
// e' di solo silenzio, FALSE altrimenti.
BOOL QVistaMicrophone::DetectSilence()
{
	UINT uFrameSize, i;
	USHORT uSample;

	return FALSE;

	uFrameSize = wfx.nChannels * wfx.wBitsPerSample / 8;

	if(uAudioBufferSize < uFrameSize){
		if(pAudioBuffer){
			delete[] pAudioBuffer;
			pAudioBuffer = NULL;
		}
		
		uAudioBufferSize = 0;
		return TRUE;
	}

	// Cicliamo un solo canale
	for(i = 0; i < uAudioBufferSize; i += 2){
		uSample = *(USHORT *)&pAudioBuffer[i];
	}

	return FALSE;
}